package edu.uiuc.edsg;

/**
 * Facade used to create iteration links for a given set of generators.
 * Static methods are used to create links from a known set of generators.
 * An instance is used to create a link iteratively.
 */
public class LinkCreator {
	
	private LinkType _type;
	private IIterationLink _curLink;
	
	public enum LinkType {
		COMBINATORIC,
		SEQUENTIAL
	}
	
	public LinkCreator(LinkType type) {
		_curLink = null;
		_type = type;
	}	

	/**
	 * Adds the given generator to the right of the iteration link tree
	 */
	public void add(IGenerator generator) {
		add(new GeneratorLink(generator));
	}
	
	/**
	 * Adds the given collection of generators to the iteration link tree
	 */
	public void addAll(Iterable<? extends IGenerator> generators) {
		for (IGenerator gen : generators) {
			add(gen);
		}
	}
	
	/**
	 * Adds the given link to the right of the iteration link tree
	 */
	public void add(IIterationLink link) {
		if (_curLink == null) {
			_curLink = link;
		}
		else if (_type == LinkType.COMBINATORIC) {
			_curLink = new CombinatoricLink(_curLink, link);	
		}
		else if (_type == LinkType.SEQUENTIAL) {
			_curLink = new SequentialLink(_curLink, link);	
		}
		else {
			throw new IllegalArgumentException("Unrecognized link type");
		}
	}
	
	/**
	 * Returns the root of the current iteration link tree 
	 */
	public IIterationLink getLink() {		
		return _curLink;
	}
	
	/**
	 * Creates a Link that when iterated will exhaustively iterate all the given generators
	 * @param generators The generators to iterate
	 * @return the link that will exhaustively iterate the given generators
	 */
	public static IIterationLink createCombinatoricLink(IGenerator... generators) {
		if (generators.length == 0) {
			throw new IllegalArgumentException();
		}
		
		IIterationLink returnLink = new GeneratorLink(generators[0]);
		for (int i = 1; i < generators.length; i++) {	
			returnLink = new CombinatoricLink(returnLink, new GeneratorLink(generators[i]));
		}
		
		return returnLink;		
	}
	
	/**
	 * Creates a Link that when iterated will iterate such that all generators have iterated to completion at least once
	 * @param generators The generators to iterate
	 * @return the link that will iterate the given generators
	 */
	public static IIterationLink createSequentialLink(IGenerator... generators) {
		if (generators.length == 0) {
			throw new IllegalArgumentException();
		}
		
		IIterationLink returnLink = new GeneratorLink(generators[0]);
		for (int i = 1; i < generators.length; i++) {
			returnLink = new SequentialLink(returnLink, new GeneratorLink(generators[i]));
		}
		
		return returnLink;		
	}

	/**
	 * Creates a Link that will iterate the given generators in the manner specified by the expression.
	 * The expression acts like a format string with the following terms:
	 * <ul>
	 *  <li>$n corresponds to the nth generator in the list (zero-based). Note that a particular generator can only be used once in an expression.</li>
	 *  <li>+ links two generators sequentially (that is, the expression will iterate as long as one of the generators has not yet completed).</li>
	 *  <li>* links two generators combinatorically.</li>
	 *  <li>( and ) affect how expressions are linked</li>
	 * </ul>
	 * Order of operations follows normal arithmetic operators. 
	 * @param expression The expression specifying how generators are linked
	 * @param generators The generators to link
	 * @return A link that will iterate the given generators as specified by the expression
	 */
	public static IIterationLink createLinkFromExpression(
			String expression, 
			IGenerator... generators) throws LinkExpressionParseException {
		return LinkExpressionParser.parse(expression, generators);
	}
	


}
