package edu.uiuc.edsg;

/**
 * Generator whose value depends on one or more other generators.
 * The wrapper acts like a literal, returning some value that depends
 *  on the current value(s) of wrapped generator(s).
 * Note that the wrapped generator(s) should be iterated before this
 *  wrapper.
 *
 * @param <ValueType> The type of value that the wrapper returns
 */
public abstract class Wrapper<ValueType> implements IGenerator<ValueType> {
	private final IGenerator[] _dependentGenerators;
	private boolean _isReset;
	
	public Wrapper(final IGenerator... dependentGenerators) {
		_isReset = true;
		_dependentGenerators = dependentGenerators;
	}	
	
	/**
	 * Gets the value from the derived class after checking that the wrapped generators
	 *  have already been iterated.
	 */
	private ValueType getDependentValue_WithIterationCheck() {
		for (IGenerator gen : _dependentGenerators) {
			if (gen.isReset()) {
				throw new IterationException("Wrapped generator must be iterated before wrapper");
			}
		}		
		return getDependentValue();
	}
	
	/**
	 * Returns the value from the derived wrapper based on the value(s) of the wrapped generator(s)
	 */
	protected abstract ValueType getDependentValue();
	
	/**
	 * {@inheritDoc}
	 */
	public ValueType current() {
		// Note: can't cache current value because the inner generator's value might change
		return getDependentValue_WithIterationCheck();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean hasNext() {
		return isReset();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean isReset() {
		return _isReset;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public ForLoopIterator<ValueType> iterator() {
		return new ForLoopIterator<ValueType>(this);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void reset() {
		_isReset = true;
	}

	/**
	 * {@inheritDoc}
	 */
	public ValueType next() {		
		_isReset = false;
		// Note: can't cache current value because the inner generator's value might change
		return getDependentValue_WithIterationCheck();
	}
}
