package edu.uiuc.edsg.test;

import org.junit.*;
import java.util.Hashtable;

import edu.uiuc.edsg.*;

public class GeneratorTestUtil {

	/**
	 * Utility function used to test that a given generator's basic functionality 
	 * works correctly.
	 * Tests reset(), isReset(), next(), hasnext(), and current()
	 * 
	 * @param generator The generator to test
	 * @param tester The <code>IGeneratedObjectTester</code> used to verify the generated objects
	 */
	public static <GeneratedObjectType> void testBasicFunctionality(
			IGenerator<GeneratedObjectType> generator,			
			IGeneratedObjectTester<GeneratedObjectType> tester) {
		int generatedCount;
		int runCount = 2; // loop twice to test reset()		
		GeneratedObjectType generated;		
		
		System.out.println("Testing " + generator.getClass().toString());
		while (runCount-- > 0) {
			generatedCount = -1;
			tester.reset();			
			generator.reset();
			Assert.assertTrue(generator.isReset()); // did you remember to call super.reset()?
			if (runCount == 0) {
				System.out.println("After Reset:");
			}
			while (generator.hasNext()) {
				Assert.assertTrue(generator.hasNext()); // make sure there are no value caching problems
				generated = generator.next();
				generatedCount++;				
				
				System.out.println("Generated Object #" + generatedCount);
				if (generated == null) {
					System.out.println("<NULL>");
				}
				else {
					System.out.println(generated.toString());
				}					
				
				Assert.assertTrue("current() does not match previous next()", tester.areEqual(generated, generator.current()));
				Assert.assertTrue(String.format("Unexpected or duplicate object found: \"%0$s\"", generated), tester.wasExpected(generated));				
			}
			Assert.assertTrue("Not all expected objects were generated", tester.foundAllExpected());
		}
	}
	
	
	/**
	 * Utility function used to test that a given generator's basic functionality 
	 * works correctly.
	 * Tests reset(), isReset(), next(), hasnext(), and current()
	 * 
	 * @param generator The generator to test
	 * @param expectedCount The number of objects the generator is expected to generate
	 */
	public static void testBasicFunctionality(
			IGenerator generator,			
			final int expectedCount) {
		
		testBasicFunctionality(
				(IGenerator<Object>)generator, 
				new IGeneratedObjectTester<Object>() {
					final int _expectedCount = expectedCount;
					int _actualCount = 0;
					
					public boolean areEqual(Object expected, Object actual) {						
						return (expected == null && actual == null) || expected.equals(actual);
					}
					public boolean foundAllExpected() {
						return _actualCount == _expectedCount;
					}
					public void reset() {
						_actualCount = 0;
					}
					public boolean wasExpected(Object generatedObject) {
						_actualCount++;
						return true;
					}
				});
	}	
	
	/**
	 * Utility function used to test that a given generator's basic functionality 
	 * works correctly.
	 * Tests reset(), isReset(), next(), hasnext(), and current()
	 * 
	 * @param generator The generator to test
	 * @param expectedObjects The objects the generator is expected to generate (Note that the class should override <code>equals</code>)
	 */
	public static <GeneratedObjectType> void testBasicFunctionality(
			IGenerator<GeneratedObjectType> generator,			
			final GeneratedObjectType... expectedObjects) {
		
		testBasicFunctionality(
				generator, 
				new IGeneratedObjectTester<GeneratedObjectType>() {
					Hashtable<GeneratedObjectType, Boolean> _found;
					
					{
						_found = new Hashtable<GeneratedObjectType, Boolean>(expectedObjects.length);
						for (GeneratedObjectType expectedObject : expectedObjects) {
							_found.put(expectedObject, false);
						}
					}
					
					public boolean areEqual(GeneratedObjectType expected, GeneratedObjectType actual) {
						return (expected == null && actual == null) || expected.equals(actual);
					}
					public boolean foundAllExpected() {						
						for (Boolean isFound : _found.values()) {
							if (!isFound) return false;
						}
						return true;
					}
					public void reset() {
						for (GeneratedObjectType key : _found.keySet()) {
							_found.put(key, false);
						}
					}					
					public boolean wasExpected(GeneratedObjectType generatedObject) {
						// if value is expected and has not already been found
						if (_found.containsKey(generatedObject) && !_found.get(generatedObject)) {
							_found.put(generatedObject, true);
							return true;
						}
						return false;
					}
				});
	}
	
	/**
	 * Utility function used to test that a given generator's basic functionality 
	 * works correctly.
	 * Tests reset(), isReset(), next(), hasnext(), and current()
	 * 
	 * @param generator The generator to test
	 * @param expectedStringValues The expected values returned from calling <code>toString</code> on the generated objects
	 */
	public static void testBasicFunctionality(
			IGenerator generator,			
			final String... expectedStringValues) {
		testBasicFunctionality(
				(IGenerator<Object>)generator, 
				new IGeneratedObjectTester<Object>() {
					Hashtable<String, Boolean> _found;
					
					{
						_found = new Hashtable<String, Boolean>(expectedStringValues.length);
						for (String expectedString : expectedStringValues) {
							_found.put(expectedString, false);
						}
					}
					
					public boolean areEqual(Object expected, Object actual) {
						return (expected == null && actual == null) || expected.equals(actual);
					}
					public boolean foundAllExpected() {						
						for (Boolean isFound : _found.values()) {
							if (!isFound) return false;
						}
						return true;
					}
					public void reset() {
						for (String key : _found.keySet()) {
							_found.put(key, false);
						}
					}					
					public boolean wasExpected(Object generatedObject) {	
						String stringVal = generatedObject.toString();
						// if value is expected and has not already been found
						if (_found.containsKey(stringVal) && !_found.get(stringVal)) {
							_found.put(stringVal, true);
							return true;
						}
						return false;
					}
				});
		
	}
	
	/**
	 * Utility function used to test that a given generator's basic functionality 
	 * works correctly.
	 * Tests reset(), isReset(), next(), hasnext(), and current()
	 * 
	 * @param generator The generator to test
	 * @param expected The expected objects and their indices (does not need to be exhaustive) 
	 */
	public static void testBasicFunctionality(
			IGenerator generator,			
			final ObjectIndexPair... expected) {
		testBasicFunctionality(
				(IGenerator<Object>)generator, 
				new IGeneratedObjectTester<Object>() {
					Hashtable<Long, Object> _expectedObjects;					
					long _curIndex;					
					
					{
						reset();
					}
										
					public boolean areEqual(Object expected, Object actual) {
						return (expected == null && actual == null) || expected.equals(actual);
					}
					public boolean foundAllExpected() {			
						// objects are removed as they are found
						return _expectedObjects.size() == 0;
					}
					public void reset() {
						_curIndex = -1;
						_expectedObjects = new Hashtable<Long, Object>();
						for (ObjectIndexPair pair : expected) {
							if (_expectedObjects.containsKey(pair.getExpectedIndex())) {
								throw new IllegalArgumentException("Duplicate index");
							}
							_expectedObjects.put(pair.getExpectedIndex(), pair.getExpectedObject());
						}
					}		
					public boolean wasExpected(Object generatedObject) {
						_curIndex++;
						
						// if the current index is one of the expected indices
						if (_expectedObjects.containsKey(_curIndex)) {
							// and the object at the index matches the given object
							if (areEqual(_expectedObjects.get(_curIndex), generatedObject)) {
								// remove the expected item
								_expectedObjects.remove(_curIndex);
								return true;
							}
							// else the generated object did not match the expected object at the index
							return false;
						}					
						// else we don't care about the generated object 
						return true;
					}
				});
		
	}
	
	@Test()
	public void expectedObjectCount() {
		Chain<Character> seq = new Chain<Character>('a','b','c');
		testBasicFunctionality(seq, 3);		
	}
	
	@Test(expected=java.lang.AssertionError.class)
	public void expectedObjectCount_InvalidCount() {
		Chain<Character> seq = new Chain<Character>('a','b','c');
		testBasicFunctionality(seq, 200);		
	}
		
	@Test()
	public void expectedObjectArray() {
		final Character[] vals = new Character[] {'a','b','c'};
		Chain<Character> seq = new Chain<Character>(vals);
		testBasicFunctionality(seq, vals);		
	}
	
	@Test(expected=java.lang.AssertionError.class)
	public void expectedObjectArray_GenerateTooFew() {
		final Character[] vals = new Character[] {'a','b','c'};
		Chain<Character> seq = new Chain<Character>(vals);
		testBasicFunctionality(seq, 'a', 'b');		
	}
	
	@Test(expected=java.lang.AssertionError.class)
	public void expectedObjectArray_GenerateTooMany() {
		final Character[] vals = new Character[] {'a','b','c'};
		Chain<Character> seq = new Chain<Character>(vals);
		testBasicFunctionality(seq, 'a','b','c','d');		
	}
	
	@Test()
	public void expectedStringValues() {
		final Integer[] vals = new Integer[] {1,2,3};
		final String[] stringVals = new String[] {"1","2","3"};
		
		Chain<Integer> seq = new Chain<Integer>(vals);
		testBasicFunctionality(seq, stringVals);		
	}
	
	@Test(expected=java.lang.AssertionError.class)
	public void expectedStringValues_GenerateTooFew() {
		final Integer[] vals = new Integer[] {1,2,3};
		final String[] stringVals = new String[] {"1","2"};
		
		Chain<Integer> seq = new Chain<Integer>(vals);
		testBasicFunctionality(seq, stringVals);		
	}
	
	@Test(expected=java.lang.AssertionError.class)
	public void expectedStringValues_GenerateTooMany() {
		final Integer[] vals = new Integer[] {1,2,3};
		final String[] stringVals = new String[] {"1","2","3","4"};
		
		Chain<Integer> seq = new Chain<Integer>(vals);
		testBasicFunctionality(seq, stringVals);		
	}
	
	@Test() 
	public void expectedAtIndex() {
		final Character[] vals = new Character[] {'a', 'b', 'c'};
		
		Chain<Character> seq = new Chain<Character>(vals);
		testBasicFunctionality(seq, new ObjectIndexPair(vals[1], 1));
	}
	
	@Test() 
	public void expectedAtIndices() {
		final Character[] vals = new Character[] {'a', 'b', 'c', 'd', 'e'};
		
		Chain<Character> seq = new Chain<Character>(vals);
		testBasicFunctionality(seq, new ObjectIndexPair(vals[1], 1), new ObjectIndexPair(vals[3], 3));
	}
	
	@Test(expected=AssertionError.class) 
	public void expectedAtIndices_NonexistentIndex() {
		final Character[] vals = new Character[] {'a', 'b', 'c', 'd', 'e'};
		
		Chain<Character> seq = new Chain<Character>(vals);
		testBasicFunctionality(seq, new ObjectIndexPair(vals[1], 1), new ObjectIndexPair(vals[3], 50));
	}
	
	@Test(expected=IllegalArgumentException.class) 
	public void expectedAtIndices_InvalidIndex() {
		final Character[] vals = new Character[] {'a', 'b', 'c', 'd', 'e'};
		
		Chain<Character> seq = new Chain<Character>(vals);
		testBasicFunctionality(seq, new ObjectIndexPair(vals[1], 1), new ObjectIndexPair(vals[3], -2));
	}
	
	@Test(expected=IllegalArgumentException.class) 
	public void expectedAtIndices_DuplicateIndex() {
		final Character[] vals = new Character[] {'a', 'b', 'c', 'd', 'e'};
		
		Chain<Character> seq = new Chain<Character>(vals);
		testBasicFunctionality(seq, new ObjectIndexPair(vals[1], 1), new ObjectIndexPair(vals[3], 1));
	}
}
