package edu.uiuc.edsg.test;

import org.junit.*;

import edu.uiuc.edsg.*;
import edu.uiuc.edsg.samples.Pair;
import edu.uiuc.edsg.samples.PairGenerator;

public class WrapperReorderVisitorTest {
	Chain<Integer> targetGenerator = new Chain<Integer>(1, 2);
	static final int TARGET_COUNT = 2;	
	SimpleWrapper<Integer> wrapper = new SimpleWrapper<Integer>(targetGenerator);
	// Wrapper count will always be 1 because it acts like a literal
	Chain<Character> otherGenerator = new Chain<Character>('a', 'b');
	static final int OTHER1_COUNT = 2;
	Chain<Character> otherGenerator2 = new Chain<Character>('c', 'd', 'e');
	static final int OTHER2_COUNT = 3;
	Chain<Character> otherGenerator3 = new Chain<Character>('f', 'g', 'h', 'i');
	static final int OTHER3_COUNT = 4;
	Chain<Character> otherGenerator4 = new Chain<Character>('j', 'k');
	static final int OTHER4_COUNT = 2;
	
	@Before()
	public void setup() {
		targetGenerator.reset();
		wrapper.reset();
		otherGenerator.reset();
		otherGenerator2.reset();
		otherGenerator3.reset();
		otherGenerator4.reset();
	}	
	
	@Test() 
	public void combinatoricLink_CorrectOrder() {		
		correctOrder(LinkCreator.LinkType.COMBINATORIC);			
	}
	@Test() 
	public void sequentialLink_CorrectOrder() {		
		correctOrder(LinkCreator.LinkType.SEQUENTIAL);			
	}
	private void correctOrder(LinkCreator.LinkType linkType) {
		LinkCreator lc = new LinkCreator(linkType);
		// wrapper is on left and should not move
		lc.add(wrapper);
		lc.add(targetGenerator);
		BinaryLink link = (BinaryLink)lc.getLink();
		
		// reordering uses CombinatoricLink regardless of original link type
		CombinatoricLink reordered = (CombinatoricLink)new WrapperReorderVisitor().getReorderedIterator(link);
		
		// check that wrapper did not move
		Assert.assertEquals(link.getLeft(), reordered.getLeft());		
		Assert.assertEquals(link.getRight(), reordered.getRight());
		
		// make sure iteration works
		testIteration(reordered, TARGET_COUNT);
	}
	
	@Test() 
	public void combinatoricLink_Reversed() {	
		reversed(LinkCreator.LinkType.COMBINATORIC);		
	}
	@Test() 
	public void sequentialLink_Reversed() {	
		reversed(LinkCreator.LinkType.SEQUENTIAL);		
	}
	private void reversed(LinkCreator.LinkType linkType) {
		LinkCreator lc = new LinkCreator(linkType);
		lc.add(targetGenerator);
		// wrapper is on right and should move left
		lc.add(wrapper);
		BinaryLink link = (BinaryLink)lc.getLink();
		
		// reordering uses CombinatoricLink regardless of original link type
		CombinatoricLink reordered = (CombinatoricLink)new WrapperReorderVisitor().getReorderedIterator(link);
		
		// check that order reversed
		Assert.assertEquals(link.getLeft(), reordered.getRight());		
		Assert.assertEquals(link.getRight(), reordered.getLeft());
		
		// test that iteration works
		testIteration(reordered, TARGET_COUNT);
	}
		
	@Test() 
	public void combinatoricLink_DeepTree() {				
		deepTree(LinkCreator.LinkType.COMBINATORIC, OTHER1_COUNT * OTHER2_COUNT * TARGET_COUNT * OTHER3_COUNT);
	}
	@Test() 
	public void sequentialLink_DeepTree() {				
		deepTree(LinkCreator.LinkType.SEQUENTIAL, Math.max(Math.max(OTHER1_COUNT, OTHER2_COUNT), Math.max(TARGET_COUNT, OTHER3_COUNT)));
	}	
	private void deepTree(LinkCreator.LinkType linkType, int expectedCount) {
		LinkCreator lc = new LinkCreator(linkType);
		lc.add(otherGenerator);
		lc.add(targetGenerator);
		lc.add(otherGenerator2);
		lc.add(wrapper);
		lc.add(otherGenerator3);
		IIterationLink link = lc.getLink();
		
		// create reordered tree (wrapper should move left)
		IIterationLink reordered = new WrapperReorderVisitor().getReorderedIterator(link);

		// make sure the original link has the correct order
		new ExpectedGeneratorsVisitor().assertFoundAll(link, 
				otherGenerator, 
				targetGenerator, 
				otherGenerator2, 
				wrapper, 
				otherGenerator3);
		// make sure the reordered link has the correct order
		new ExpectedGeneratorsVisitor().assertFoundAll(reordered , 
				wrapper, // note that order has changed
				otherGenerator, 				
				targetGenerator, 
				otherGenerator2, 
				otherGenerator3);
		
		// check iteration
		testIteration(reordered, expectedCount);		
	}
	
	@Test()
	public void multipleWrappers() {
		SimpleWrapper<Integer> wrapper2 = new SimpleWrapper<Integer>(targetGenerator);
		
		IIterationLink link = LinkCreator.createCombinatoricLink(
				otherGenerator, 
				targetGenerator, 
				otherGenerator2, 
				wrapper, 
				otherGenerator3, 
				wrapper2, 
				otherGenerator4);
		
		// wrappers should move to front of list, retaining their relative order
		IIterationLink reordered = new WrapperReorderVisitor().getReorderedIterator(link);
		
		// make sure the original link has the correct order
		new ExpectedGeneratorsVisitor().assertFoundAll(link, 
				otherGenerator, 
				targetGenerator, 
				otherGenerator2, 
				wrapper, 
				otherGenerator3, 
				wrapper2, 
				otherGenerator4);
		// make sure the reordered link has the correct order
		new ExpectedGeneratorsVisitor().assertFoundAll(reordered , 
				wrapper, 
				wrapper2, 
				otherGenerator, 
				targetGenerator, 
				otherGenerator2, 				
				otherGenerator3, 				
				otherGenerator4);
		
		testIteration(reordered, OTHER1_COUNT * TARGET_COUNT * OTHER2_COUNT  * OTHER3_COUNT  * OTHER4_COUNT);
	}
	
	@Test @Ignore
	public void wrappingWrapper() {
		
		
		// NOTE: KNOWN TEST FAILURE  (Brett 12/5/06)
		
		
		SimpleWrapper<Integer> wrapperWrapper = new SimpleWrapper<Integer>(wrapper);
		// note that wrapper/wrapperWrapper ordering is reversed of what it should be
		IIterationLink link = LinkCreator.createCombinatoricLink(
				otherGenerator, 
				wrapper, 
				otherGenerator2, 
				wrapperWrapper, 
				otherGenerator3, 
				targetGenerator);	

		IIterationLink reordered = new WrapperReorderVisitor().getReorderedIterator(link);

		// make sure the reordered link has the correct order
		new ExpectedGeneratorsVisitor().assertFoundAll(reordered , 
				// wrapper should be iterated before wrapperWrapper
				wrapperWrapper, 
				wrapper, 
				otherGenerator, 				
				otherGenerator2, 
				otherGenerator3, 
				targetGenerator);
		
		testIteration(reordered, OTHER1_COUNT * OTHER2_COUNT * OTHER3_COUNT * TARGET_COUNT);		
	}
	
	/**
	 * Tests that CompositeGenerators are expanded into their iteration trees
	 */
	@Test()
	public void treeComposition() {
		CompositeGenerator<Pair<Character, Character>> pair1Gen = 
			new PairGenerator<Character, Character>(otherGenerator, otherGenerator2);
		CompositeGenerator<Pair<Character, Character>> pair2Gen = 
			new PairGenerator<Character, Character>(otherGenerator3, otherGenerator4);
		
		IIterationLink originalTree = LinkCreator.createCombinatoricLink(pair1Gen, pair2Gen);
		IIterationLink fullTree = new WrapperReorderVisitor().getReorderedIterator(originalTree);
		
		// original tree will only contain two generators
		new ExpectedGeneratorsVisitor().assertFoundAll(originalTree, 
				pair1Gen, 
				pair2Gen);
		// full tree will have "expanded" the generators into their iteration trees
		new ExpectedGeneratorsVisitor().assertFoundAll(fullTree, 
				otherGenerator, 
				otherGenerator2, 
				otherGenerator3, 
				otherGenerator4);
		
		testIteration(fullTree, OTHER1_COUNT * OTHER2_COUNT * OTHER3_COUNT * OTHER4_COUNT);
	}
	
	/**
	 * Tests that composite generators are expanded into their iteration trees and 
	 * that wrappers contained in the composite generators are moved to the end of
	 * the iteration.
	 */
	@Test()
	public void treeComposition_WithWrappers() {
		CompositeGenerator<Pair<Character, Character>> pair1Gen = 
			new PairGenerator<Character, Character>(otherGenerator, otherGenerator2);
		CompositeGenerator<Pair<Integer, Integer>> pair2Gen = 
			new PairGenerator<Integer, Integer>(targetGenerator, wrapper);
		
		IIterationLink originalTree = LinkCreator.createCombinatoricLink(pair1Gen, pair2Gen);
		IIterationLink fullTree = new WrapperReorderVisitor().getReorderedIterator(originalTree);
		
		// original tree will only contain two generators
		new ExpectedGeneratorsVisitor().assertFoundAll(originalTree, 
				pair1Gen, 
				pair2Gen);
		// full tree will have "expanded" the generators into their iteration trees
		// and moved wrapper to end of iteration
		new ExpectedGeneratorsVisitor().assertFoundAll(fullTree, 
				wrapper, 
				otherGenerator, 
				otherGenerator2, 
				targetGenerator);

		testIteration(fullTree, OTHER1_COUNT * OTHER2_COUNT * TARGET_COUNT);
	}
	
	/**
	 * Utility function to test that the generators are behaving correctly
	 */
	private void testIteration(IIterationLink link, int expectedCount) {
		int count = 0;
		for (link.reset(); link.hasNext(); link.next()) {
			count++;
		}
		Assert.assertEquals(expectedCount, count);
	}	
}
